@@ -9,11 +9,12 @@
 use git2::{self, Config as GitConfig, Error as GitError, Cred as GitCred, RemoteCallbacks, CredentialType, FetchOptions, ProxyOptions, Repository, Tree, Oid};
 use semver::{VersionReq as SemverReq, Version as Semver};
 use std::io::{ErrorKind as IoErrorKind, Write, Read};
+use std::ffi::{OsString, OsStr};
 use std::collections::BTreeMap;
 use std::path::{PathBuf, Path};
 use std::hash::{Hasher, Hash};
-use std::fs::{self, File};
-use std::{cmp, env, mem};
+use std::{cmp, env, mem, fs};
+use std::process::Command;
 use std::borrow::Cow;
 use regex::Regex;
 use url::Url;
@@ -466,38 +467,54 @@ impl GitRepoPackage {
     }
 
     /// Clone the repo and check what the latest commit's hash is.
-    pub fn pull_version<Pt: AsRef<Path>, Pg: AsRef<Path>>(&mut self, temp_dir: Pt, git_db_dir: Pg, http_proxy: Option<&str>) {
-        self.pull_version_impl(temp_dir.as_ref(), git_db_dir.as_ref(), http_proxy)
+    pub fn pull_version<Pt: AsRef<Path>, Pg: AsRef<Path>>(&mut self, temp_dir: Pt, git_db_dir: Pg, http_proxy: Option<&str>, fork_git: bool) {
+        self.pull_version_impl(temp_dir.as_ref(), git_db_dir.as_ref(), http_proxy, fork_git)
     }
 
-    fn pull_version_impl(&mut self, temp_dir: &Path, git_db_dir: &Path, http_proxy: Option<&str>) {
+    fn pull_version_impl(&mut self, temp_dir: &Path, git_db_dir: &Path, http_proxy: Option<&str>, fork_git: bool) {
         let clone_dir = find_git_db_repo(git_db_dir, &self.url).unwrap_or_else(|| {
             fs::create_dir_all(temp_dir).unwrap();
             temp_dir.join(&self.name)
         });
 
-        let repo = self.pull_version_repo(&clone_dir, http_proxy);
+        let repo = self.pull_version_repo(&clone_dir, http_proxy, fork_git);
 
         self.newest_id = Some(repo.and_then(|r| r.head().and_then(|h| h.target().ok_or_else(|| GitError::from_str("HEAD not a direct reference")))).unwrap());
     }
 
-    fn pull_version_fresh_clone(&self, clone_dir: &Path, http_proxy: Option<&str>) -> Result<Repository, GitError> {
-        with_authentication(&self.url, |creds| {
-            let mut bldr = git2::build::RepoBuilder::new();
+    fn pull_version_fresh_clone(&self, clone_dir: &Path, http_proxy: Option<&str>, fork_git: bool) -> Result<Repository, GitError> {
+        if fork_git {
+            Command::new(env::var_os("GIT").as_ref().map(OsString::as_os_str).unwrap_or(OsStr::new("git")))
+                .arg("clone")
+                .args(self.branch.as_ref().map(|_| "-b"))
+                .args(self.branch.as_ref())
+                .args(&["--bare", "--", &self.url])
+                .arg(clone_dir)
+                .status()
+                .map_err(|e| GitError::from_str(&e.to_string()))
+                .and_then(|e| if e.success() {
+                    Repository::open(clone_dir)
+                } else {
+                    Err(GitError::from_str(&e.to_string()))
+                })
+        } else {
+            with_authentication(&self.url, |creds| {
+                let mut bldr = git2::build::RepoBuilder::new();
 
-            let mut cb = RemoteCallbacks::new();
-            cb.credentials(|a, b, c| creds(a, b, c));
-            bldr.fetch_options(fetch_options_from_proxy_url_and_callbacks(&self.url, http_proxy, cb));
-            if let Some(ref b) = self.branch.as_ref() {
-                bldr.branch(b);
-            }
+                let mut cb = RemoteCallbacks::new();
+                cb.credentials(|a, b, c| creds(a, b, c));
+                bldr.fetch_options(fetch_options_from_proxy_url_and_callbacks(&self.url, http_proxy, cb));
+                if let Some(ref b) = self.branch.as_ref() {
+                    bldr.branch(b);
+                }
 
-            bldr.bare(true);
-            bldr.clone(&self.url, &clone_dir)
-        })
+                bldr.bare(true);
+                bldr.clone(&self.url, &clone_dir)
+            })
+        }
     }
 
-    fn pull_version_repo(&self, clone_dir: &Path, http_proxy: Option<&str>) -> Result<Repository, GitError> {
+    fn pull_version_repo(&self, clone_dir: &Path, http_proxy: Option<&str>, fork_git: bool) -> Result<Repository, GitError> {
         if let Ok(r) = Repository::open(clone_dir) {
             // If `Repository::open` is successful, both `clone_dir` exists *and* points to a valid repository.
             //
@@ -525,16 +542,32 @@ impl GitRepoPackage {
                             // yeeting them shouldn't be a problem, since that's what we *would* do anyway,
                             // and we set up for the non-pessimised path in later runs.
                             fs::remove_dir_all(clone_dir).unwrap();
-                            return self.pull_version_fresh_clone(clone_dir, http_proxy);
+                            return self.pull_version_fresh_clone(clone_dir, http_proxy, fork_git);
                         }
                     }
 
                 }
             };
 
+            let mut remote = "origin";
             r.find_remote("origin")
-                .or_else(|_| r.remote_anonymous(&self.url))
-                .and_then(|mut rm| {
+                .or_else(|_| {
+                    remote = &self.url;
+                    r.remote_anonymous(&self.url)
+                })
+                .and_then(|mut rm| if fork_git {
+                    Command::new(env::var_os("GIT").as_ref().map(OsString::as_os_str).unwrap_or(OsStr::new("git")))
+                        .arg("-C")
+                        .arg(r.path())
+                        .args(&["fetch", remote, &branch])
+                        .status()
+                        .map_err(|e| GitError::from_str(&e.to_string()))
+                        .and_then(|e| if e.success() {
+                            Ok(())
+                        } else {
+                            Err(GitError::from_str(&e.to_string()))
+                        })
+                } else {
                     with_authentication(&self.url, |creds| {
                         let mut cb = RemoteCallbacks::new();
                         cb.credentials(|a, b, c| creds(a, b, c));
@@ -558,13 +591,13 @@ impl GitRepoPackage {
                 .unwrap();
             Ok(r)
         } else {
-            // If we could not open the repository either it does not exist, or exists but is invalid.
+            // If we could not open the repository either it does not exist, or exists but is invalid,
+            // in which case remove it to trigger a fresh clone.
             if clone_dir.exists() {
-                // If `clone_dir` exists it must be invalid. Remove it to trigger a fresh clone.
                 fs::remove_dir_all(&clone_dir).unwrap();
             }
 
-            self.pull_version_fresh_clone(clone_dir, http_proxy)
+            self.pull_version_fresh_clone(clone_dir, http_proxy, fork_git)
         }
     }
 
@@ -652,6 +685,50 @@ impl PackageFilterElement {
 }
 
 
+/// `cargo` configuration, as obtained from `.cargo/config[.toml]`
+#[derive(Debug, Clone, Hash, PartialEq, Eq, PartialOrd, Ord)]
+pub struct CargoConfig {
+    pub net_git_fetch_with_cli: bool,
+}
+
+impl CargoConfig {
+    pub fn load(crates_file: &Path) -> CargoConfig {
+        CargoConfig {
+            net_git_fetch_with_cli: env::var("CARGO_NET_GIT_FETCH_WITH_CLI")
+                .ok()
+                .and_then(|e| if e.is_empty() {
+                    Some(toml::Value::String(String::new()))
+                } else {
+                    e.parse::<toml::Value>().ok()
+                })
+                .or_else(|| {
+                    fs::read_to_string(crates_file.with_file_name("config"))
+                        .or_else(|_| fs::read_to_string(crates_file.with_file_name("config,toml")))
+                        .ok()?
+                        .parse::<toml::Value>()
+                        .ok()?
+                        .as_table_mut()?
+                        .remove("net")?
+                        .as_table_mut()?
+                        .remove("git-fetch-with-cli")
+                })
+                .map(CargoConfig::truthy)
+                .unwrap_or(false),
+        }
+    }
+
+    fn truthy(v: toml::Value) -> bool {
+        match v {
+            toml::Value::String(ref s) if s == "" => false,
+            toml::Value::Float(f) if f == 0. => false,
+            toml::Value::Integer(0) |
+            toml::Value::Boolean(false) => false,
+            _ => true,
+        }
+    }
+}
+
+
 /// [Follow `install.root`](https://github.com/nabijaczleweli/cargo-update/issues/23) in the `config` file
 /// parallel to the specified crates file up to the final one.
 ///
@@ -665,12 +742,12 @@ impl PackageFilterElement {
 /// # let _ = crates_file;
 /// ```
 pub fn resolve_crates_file(crates_file: PathBuf) -> PathBuf {
-    let config_file = crates_file.with_file_name("config");
+    let mut config_file = crates_file.with_file_name("config");
+    if !config_file.exists() {
+        config_file.set_file_name("config.toml");
+    }
     if config_file.exists() {
-        let mut crates = String::new();
-        File::open(&config_file).unwrap().read_to_string(&mut crates).unwrap();
-
-        if let Some(idir) = toml::from_str::<toml::Value>(&crates)
+        if let Some(idir) = toml::from_str::<toml::Value>(&fs::read_to_string(&config_file).unwrap())
             .unwrap()
             .get("install")
             .and_then(|t| t.as_table())
@@ -704,11 +781,8 @@ pub fn resolve_crates_file(crates_file: PathBuf) -> PathBuf {
 /// ```
 pub fn installed_registry_packages(crates_file: &Path) -> Vec<RegistryPackage> {
     if crates_file.exists() {
-        let mut crates = String::new();
-        File::open(crates_file).unwrap().read_to_string(&mut crates).unwrap();
-
         let mut res = Vec::<RegistryPackage>::new();
-        for pkg in toml::from_str::<toml::Value>(&crates).unwrap()["v1"]
+        for pkg in toml::from_str::<toml::Value>(&fs::read_to_string(crates_file).unwrap()).unwrap()["v1"]
             .as_table()
             .unwrap()
             .iter()
@@ -749,11 +823,8 @@ pub fn installed_registry_packages(crates_file: &Path) -> Vec<RegistryPackage> {
 /// ```
 pub fn installed_git_repo_packages(crates_file: &Path) -> Vec<GitRepoPackage> {
     if crates_file.exists() {
-        let mut crates = String::new();
-        File::open(crates_file).unwrap().read_to_string(&mut crates).unwrap();
-
         let mut res = Vec::<GitRepoPackage>::new();
-        for pkg in toml::from_str::<toml::Value>(&crates).unwrap()["v1"]
+        for pkg in toml::from_str::<toml::Value>(&fs::read_to_string(crates_file).unwrap()).unwrap()["v1"]
             .as_table()
             .unwrap()
             .iter()
@@ -789,13 +860,18 @@ pub fn installed_git_repo_packages(crates_file: &Path) -> Vec<GitRepoPackage> {
 /// #                            "registry+https://github.com/rust-lang/crates.io-index".to_string())];
 /// let mut installed_packages = installed_registry_packages(&cargo_dir);
 /// # let mut installed_packages =
-/// #     vec![RegistryPackage::parse("cargo-outdated 0.2.0 (registry+https://github.com/rust-lang/crates.io-index)", vec!["cargo-outdated".to_string()]).unwrap(),
-/// #          RegistryPackage::parse("racer 1.2.10 (registry+https://github.com/rust-lang/crates.io-index)", vec!["racer.exe".to_string()]).unwrap(),
-/// #          RegistryPackage::parse("rustfmt 0.6.2 (registry+https://github.com/rust-lang/crates.io-index)", vec!["rustfmt".to_string(), "cargo-format".to_string()]).unwrap()];
+/// #     vec![RegistryPackage::parse("cargo-outdated 0.2.0 (registry+https://github.com/rust-lang/crates.io-index)",
+/// #     vec!["cargo-outdated".to_string()]).unwrap(),
+/// #          RegistryPackage::parse("racer 1.2.10 (registry+https://github.com/rust-lang/crates.io-index)",
+/// #     vec!["racer.exe".to_string()]).unwrap(),
+/// #          RegistryPackage::parse("rustfmt 0.6.2 (registry+https://github.com/rust-lang/crates.io-index)",
+/// #     vec!["rustfmt".to_string(), "cargo-format".to_string()]).unwrap()];
 /// installed_packages = intersect_packages(&installed_packages, &packages_to_update, false, &[]);
 /// # assert_eq!(&installed_packages,
-/// #   &[RegistryPackage::parse("cargo-outdated 0.2.0 (registry+https://github.com/rust-lang/crates.io-index)", vec!["cargo-outdated".to_string()]).unwrap(),
-/// #     RegistryPackage::parse("racer 1.2.10 (registry+https://github.com/rust-lang/crates.io-index)", vec!["racer.exe".to_string()]).unwrap()]);
+/// #   &[RegistryPackage::parse("cargo-outdated 0.2.0 (registry+https://github.com/rust-lang/crates.io-index)",
+/// #                            vec!["cargo-outdated".to_string()]).unwrap(),
+/// #     RegistryPackage::parse("racer 1.2.10 (registry+https://github.com/rust-lang/crates.io-index)",
+/// #                            vec!["racer.exe".to_string()]).unwrap()]);
 /// ```
 pub fn intersect_packages(installed: &[RegistryPackage], to_update: &[(String, Option<Semver>, String)], allow_installs: bool,
                           installed_git: &[GitRepoPackage])
@@ -923,20 +999,36 @@ pub fn assert_index_path(cargo_dir: &Path, registry_url: &str) -> Result<PathBuf
 ///
 /// Most of this would have been impossible, of course, without the [`rust-lang` Discord server](https://discord.gg/rust-lang),
 /// so shoutout to whoever convinced people that Discord is actually good.
-pub fn update_index<W: Write>(index_repo: &mut Repository, repo_url: &str, http_proxy: Option<&str>, out: &mut W) -> Result<(), String> {
+///
+/// Sometimes, however, even this isn't enough (see https://github.com/nabijaczleweli/cargo-update/issues/163),
+/// hence `fork_git`, which actually runs `$GIT` (default: `git`).
+pub fn update_index<W: Write>(index_repo: &mut Repository, repo_url: &str, http_proxy: Option<&str>, fork_git: bool, out: &mut W) -> Result<(), String> {
     writeln!(out, "    Updating registry '{}'", repo_url).map_err(|_| "failed to write updating message".to_string())?;
-    index_repo.remote_anonymous(repo_url)
-        .and_then(|mut r| {
-            with_authentication(repo_url, |creds| {
-                let mut cb = RemoteCallbacks::new();
-                cb.credentials(|a, b, c| creds(a, b, c));
+    if fork_git {
+        Command::new(env::var_os("GIT").as_ref().map(OsString::as_os_str).unwrap_or(OsStr::new("git"))).arg("-C")
+            .arg(index_repo.path())
+            .args(&["fetch", "-f", repo_url, "refs/heads/master:refs/remotes/origin/master"])
+            .status()
+            .map_err(|e| e.to_string())
+            .and_then(|e| if e.success() {
+                Ok(())
+            } else {
+                Err(e.to_string())
+            })?;
+    } else {
+        index_repo.remote_anonymous(repo_url)
+            .and_then(|mut r| {
+                with_authentication(repo_url, |creds| {
+                    let mut cb = RemoteCallbacks::new();
+                    cb.credentials(|a, b, c| creds(a, b, c));
 
-                r.fetch(&["refs/heads/master:refs/remotes/origin/master"],
-                        Some(&mut fetch_options_from_proxy_url_and_callbacks(repo_url, http_proxy, cb)),
-                        None)
+                    r.fetch(&["refs/heads/master:refs/remotes/origin/master"],
+                            Some(&mut fetch_options_from_proxy_url_and_callbacks(repo_url, http_proxy, cb)),
+                            None)
+                })
             })
-        })
-        .map_err(|e| e.message().to_string())?;
+            .map_err(|e| e.message().to_string())?;
+    }
     writeln!(out).map_err(|_| "failed to write post-update newline".to_string())?;
 
     Ok(())
@@ -980,16 +1072,20 @@ fn fetch_options_from_proxy_url_and_callbacks<'a>(repo_url: &str, proxy_url: Opt
 /// Consult [#107](https://github.com/nabijaczleweli/cargo-update/issues/107) and
 /// the [Cargo Book](https://doc.rust-lang.org/cargo/reference/source-replacement.html) for details
 pub fn get_index_url(crates_file: &Path, registry: &str) -> Result<(String, Cow<'static, str>), Cow<'static, str>> {
-    let config_file = crates_file.with_file_name("config");
-    let config = if let Ok(cfg) = fs::read_to_string(&config_file) {
+    let mut config_file = crates_file.with_file_name("config");
+    let config = if let Ok(cfg) = fs::read_to_string(&config_file).or_else(|_| {
+        config_file.set_file_name("config.toml");
+        fs::read_to_string(&config_file)
+    }) {
         toml::from_str::<toml::Value>(&cfg).map_err(|e| format!("{} not TOML: {}", config_file.display(), e))?
     } else {
         if registry == "https://github.com/rust-lang/crates.io-index" {
             return Ok((registry.to_string(), "crates-io".into()));
         } else {
-            Err(format!("Non-crates.io registry specified and no config file found at {}. \
+            Err(format!("Non-crates.io registry specified and no config file found at {} or {}. \
                          Due to a Cargo limitation we will not be able to install from there \
                          until it's given a [source.NAME] in that file!",
+                        config_file.with_file_name("config").display(),
                         config_file.display()))?
         }
     };
@@ -1210,10 +1306,7 @@ pub fn find_package_data<'t>(cratename: &str, registry: &Tree<'t>, registry_pare
 pub fn find_proxy(crates_file: &Path) -> Option<String> {
     let config_file = crates_file.with_file_name("config");
     if config_file.exists() {
-        let mut crates = String::new();
-        File::open(&config_file).unwrap().read_to_string(&mut crates).unwrap();
-
-        if let Some(proxy) = toml::from_str::<toml::Value>(&crates)
+        if let Some(proxy) = toml::from_str::<toml::Value>(&fs::read_to_string(config_file).unwrap())
             .unwrap()
             .get("http")
             .and_then(|t| t.as_table())
